import { basename, dirname, extname, join } from "node:path";
import type babel from "@babel/core";
import hash from "@emotion/hash";
import { isPlainObject } from "lodash";
import invariant from "tiny-invariant";
import { type NodePath, type types as t } from "@babel/core";
import type { SourceLocation, StyleMapEntry } from "./shared";
import { type ResolveTailwindOptions, getClassName } from "./index";

const definePlugin =
  <T>(fn: (runtime: typeof babel) => babel.Visitor<babel.PluginPass & T>) =>
  (runtime: typeof babel) => {
    const plugin: babel.PluginObj<babel.PluginPass & T> = {
      visitor: fn(runtime),
    };
    return plugin as babel.PluginObj;
  };

const extractJSXContainer = (attr: NonNullable<t.JSXAttribute["value"]>): t.Expression =>
  attr.type === "JSXExpressionContainer" ? (attr.expression as t.Expression) : attr;

function matchPath(
  nodePath: NodePath<t.Node | null | undefined>,
  fns: (dig: (nodePath: NodePath<t.Node | null | undefined>) => void) => babel.Visitor
) {
  if (!nodePath.node) return;
  const fn = fns(path => matchPath(path, fns))[nodePath.node.type] as any;
  fn?.(nodePath);
}

interface BabelPluginState {
  getCx: () => t.Identifier;
  sliceText: (node: t.Node) => SourceLocation;
  recordIfAbsent: (node: StyleMapEntry) => void;
  tailwindMap: Map<string, StyleMapEntry>;
}

export type ClassNameCollector = (path: string, entries: StyleMapEntry[]) => void;

const trim = (value: string) => value.replace(/\s+/g, " ").trim();
const trimPrefix = (cls: string, prefix: string) =>
  trim(cls)
    .split(" ")
    .map(value => prefix + value);

const flatMapEntries = <K extends string | number, V, R>(
  map: Record<K, V>,
  fn: (value: V, key: K) => R[]
): R[] => Object.entries(map).flatMap(([key, value]) => fn(value as V, key as K));

export function babelTailwind(
  {
    styleMap,
    clsx,
    getClassName: getClass = getClassName,
    macroFunction,
    jsxAttributeAction = "delete",
    jsxAttributeName = "css",
    vite,
  }: ResolveTailwindOptions,
  onCollect: ClassNameCollector | undefined
) {
  function getClsxImport(t: typeof babel.types, cx: t.Identifier) {
    switch (clsx) {
      case "emotion":
        return t.importDeclaration(
          [t.importSpecifier(cx, t.identifier("cx"))],
          t.stringLiteral("@emotion/css")
        );
      case "clsx":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("clsx")
        );
      case "classnames":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("classnames")
        );
      default:
        throw new Error("Unknown clsx library");
    }
  }

  function evaluateArgs(paths: NodePath[]) {
    return paths
      .flatMap(path => {
        const { confident, value } = path.evaluate();
        invariant(confident, `${macroFunction} argument cannot be statically evaluated`);

        if (typeof value === "string") {
          return trim(value);
        }

        if (isPlainObject(value)) {
          return flatMapEntries(value, (classes, modifier) => {
            if (modifier === "data" && isPlainObject(classes)) {
              return flatMapEntries(
                classes as Record<string, string | object>,
                (cls, key) =>
                  typeof cls === "string"
                    ? trimPrefix(cls, `${modifier}-[${key}]:`)
                    : flatMapEntries(cls as Record<string, string>, (cls, attrValue) =>
                        trimPrefix(cls, `${modifier}-[${key}=${attrValue}]:`)
                      )
              );
            }

            invariant(
              typeof classes === "string",
              `Value for "${modifier}" should be a string`
            );
            return trimPrefix(classes, modifier + ":");
          });
        }

        throw new Error(`${macroFunction} argument has an invalid type`);
      })
      .join(" ");
  }

  return definePlugin<BabelPluginState>(({ types: t }) => ({
    Program: {
      enter(path, state) {
        let cx: t.Identifier;
        const map = (state.tailwindMap = new Map());
        state.sliceText = node => ({
          filename: state.filename!,
          start: node.loc!.start,
          end: node.loc!.end,
          text: state.file.code
            .split("\n")
            .slice(node.loc!.start.line - 1, node.loc!.end.line)
            .join("\n"),
        });
        state.recordIfAbsent = entry => {
          if (!map.has(entry.key)) {
            map.set(entry.key, entry);
          }
        };
        state.getCx = () => {
          if (cx == null) {
            cx = path.scope.generateUidIdentifier("cx");
            path.node.body.unshift(getClsxImport(t, cx));
          }
          return t.cloneNode(cx);
        };
      },

      exit({ node }, { filename, tailwindMap }) {
        if (!tailwindMap.size) return;
        invariant(filename, "babel: missing state.filename");

        const cssName = basename(filename, extname(filename)) + ".css";

        const path = join(dirname(filename), cssName);
        const value = Array.from(tailwindMap.values());

        let importee = `tailwind:./${cssName}`;
        if (vite) {
          const cacheKey = hash(value.map(x => x.className).join(","));
          importee += `?${cacheKey}`;
        }

        node.body.unshift(t.importDeclaration([], t.stringLiteral(importee)));

        styleMap.set(path, value);
        onCollect?.(path, value);
      },
    },

    TaggedTemplateExpression(path, { sliceText, recordIfAbsent }) {
      if (macroFunction == null) return;
      const { node } = path;

      const {
        tag,
        quasi: { quasis, expressions },
      } = node;
      if (!t.isIdentifier(tag, { name: macroFunction })) return;

      invariant(
        !expressions.length,
        `${macroFunction}\`\` should not contain expressions`
      );

      const value = quasis[0].value.cooked;
      if (value) {
        const trimmed = trim(value);
        const className = getClass(trimmed);
        recordIfAbsent({
          key: className,
          className: trimmed,
          location: sliceText(node),
        });
        path.replaceWith(t.stringLiteral(className));
      }
    },

    CallExpression(path, { sliceText, recordIfAbsent }) {
      if (macroFunction == null) return;

      const { node } = path;

      const { callee } = node;
      if (!t.isIdentifier(callee, { name: macroFunction })) return;

      const trimmed = evaluateArgs(path.get("arguments"));
      const className = getClass(trimmed);
      recordIfAbsent({
        key: className,
        className: trimmed,
        location: sliceText(node),
      });
      path.replaceWith(t.stringLiteral(className));
    },

    JSXAttribute(path, { sliceText, recordIfAbsent, getCx }) {
      const { name } = path.node;
      if (name.name !== jsxAttributeName) return;

      const valuePath = path.get("value");
      if (!valuePath.node) return;

      const copy =
        jsxAttributeAction === "delete" ? undefined : t.cloneNode(valuePath.node, true);

      const parent = path.parent as t.JSXOpeningElement;
      const classNameAttribute = parent.attributes.find(
        (attr): attr is t.JSXAttribute =>
          t.isJSXAttribute(attr) && attr.name.name === "className"
      );

      matchPath(valuePath, go => ({
        StringLiteral(path) {
          const { node } = path;
          const { value } = node;
          const trimmed = trim(value);
          if (trimmed) {
            const className = getClass(trimmed);
            recordIfAbsent({
              key: className,
              className: trimmed,
              location: sliceText(node),
            });
            path.replaceWith(t.stringLiteral(className));
          }
        },
        ArrayExpression(path) {
          for (const element of path.get("elements")) {
            go(element);
          }
        },
        ObjectExpression(path) {
          const trimmed = evaluateArgs([path]);
          const className = getClass(trimmed);
          recordIfAbsent({
            key: className,
            className: trimmed,
            location: sliceText(path.node),
          });
          path.replaceWith(t.stringLiteral(className));
        },
        JSXExpressionContainer(path) {
          go(path.get("expression"));
        },
        ConditionalExpression(path) {
          go(path.get("consequent"));
          go(path.get("alternate"));
        },
        LogicalExpression(path) {
          go(path.get("right"));
        },
        CallExpression(path) {
          for (const arg of path.get("arguments")) {
            go(arg);
          }
        },
      }));

      let valuePathNode = extractJSXContainer(valuePath.node);
      if (
        t.isArrayExpression(valuePathNode) &&
        valuePathNode.elements.every(node => t.isStringLiteral(node))
      ) {
        valuePathNode = t.stringLiteral(
          // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
          (valuePathNode.elements as t.StringLiteral[]).map(node => node.value).join(" ")
        );
      }

      if (classNameAttribute) {
        const attrValue = classNameAttribute.value!;
        const wrap = (originalValue: babel.types.Expression) =>
          t.callExpression(getCx(), [originalValue, valuePathNode]);

        // If both are string literals, we can merge them directly here
        if (t.isStringLiteral(attrValue) && t.isStringLiteral(valuePathNode)) {
          attrValue.value +=
            (attrValue.value.at(-1) === " " ? "" : " ") + valuePathNode.value;
        } else {
          const internalAttrValue = extractJSXContainer(attrValue);
          if (
            t.isArrowFunctionExpression(internalAttrValue) &&
            !t.isBlockStatement(internalAttrValue.body)
          ) {
            internalAttrValue.body = wrap(internalAttrValue.body);
          } else {
            classNameAttribute.value = t.jsxExpressionContainer(wrap(internalAttrValue));
          }
        }
      } else {
        parent.attributes.push(
          t.jsxAttribute(
            t.jsxIdentifier("className"),
            t.jSXExpressionContainer(valuePathNode!)
          )
        );
      }

      if (jsxAttributeAction === "delete") {
        path.remove();
      } else {
        path.node.value = copy!;
        if (Array.isArray(jsxAttributeAction) && jsxAttributeAction[0] === "rename") {
          path.node.name.name = jsxAttributeAction[1];
        }
      }
    },
  }));
}
